


######################################################################
###  NUMERICKE METODY NEPODMINENE MINIMALIZACE V R^n - ZDROJOVY KOD
###  © Petr ZEMANEK
###  URCENO PRO POTREBY PREDMETU M5170
###  VERZE: 7. dubna 2017
######################################################################
MetodyVRn:=module()

description "Numericke metody nepodminene minimalizace v R^n (Metody nejvetsiho spadu /MNS/, Newtonova metoda /NM/, Metoda sdruzenych gradientu /MSG/, Metoda sdruzenych gradientu s resetem hodnoty beta /MSGreset/";  

option package;  

uses Optimization, Student[MultivariateCalculus], Student[VectorCalculus], LinearAlgebra;  

export MNS, NM, MSG, MSGreset;

############################
############################

MNS := proc (f, P, N, VypsatIterace,VypsatIteracePodrobne)

description "MSN=metoda nejvetsiho spadu, obecna syntaxe MNS(f, P, N, VypsatIterace,VypsatIteracePodrobne), f..funkce, P..pocet promennych (ty musi byt zapsany jako x[1], x[2],...), N..pocet iteraci, VypsatIterace=true/false, VypsatIteracePodrobne=true/false";

local F, n, x0, Gradienty, i, Krok, a,b, newx0, Smer, Derivace, Promenne, HodnotaPromennych, j, VycisleniDerivace, c, FunkceMinSearch, MinSearch, Iterace, VychoziHodnota, FunkcniHodnota;

for i to P do
 x0[i]:=VychoziBod[i];
end do;

Iterace[0]:=convert(x0,list);

for i to P do
 a[i]:=x[i];
 b[i]:=x[i]=x0[i]; 
end do;

Promenne:=convert(a,list);
HodnotaPromennych:=convert(b,list);

VychoziHodnota:=eval(f,HodnotaPromennych);

for i to P do 
 Derivace[i]:=diff(f,x[i]);
end do; 

print('__________________________________________');
print("FUNKCE", f); 
print("PRESNE RESENI", minimize(f,location)); 
print("METODA NEJVETSIHO SPADU");
print('__________________________________________');

print("VYCHOZI BOD", HodnotaPromennych,"S HODNOTOU", simplify(VychoziHodnota));
print('__________________________________________');

print('GRADIENT',PositionVector(convert(Derivace,list))); 
print('__________________________________________');
 
if VypsatIteracePodrobne
 then print("k","FUNKCE PRO VYPOCET ALPHA", "ALPHA_(k-1)", "[VEKTOR h_(k-1)]", "[X_k=X_(k-1)+ALPHA_(k-1)+h_(k-1)]",  "F(X_k)");
end if; 

if VypsatIterace
 then print("k", "ALPHA_(k-1)", "[VEKTOR h_(k-1)]", "[X_k=X_(k-1)+ALPHA_(k-1)+h_(k-1)]", "F(X_k)");
end if; 

print('__________________________________________');

F:=unapply(f,promenne);

for i to N while evalf(eval(sum(Derivace[k]*Derivace[k],k=1..P),HodnotaPromennych))>0 do

 for j to P do
  VycisleniDerivace[j]:=eval(Derivace[j],HodnotaPromennych);
  c[j]:=x[j]=x0[j]-alpha*VycisleniDerivace[j];
 end do;

 Gradienty:=convert(c,list);
 Smer:=convert(VycisleniDerivace,list);

  FunkceMinSearch:=eval(f,Gradienty);

 MinSearch:=minimize(FunkceMinSearch,alpha=0..infinity,location):

 if whattype(MinSearch[1])=function 
  then 
    try
             MinSearch:=Minimize(FunkceMinSearch,{alpha>=0}):
             Krok := identify((subs(indets(MinSearch[2], `=`), alpha)),all=false,Rational=true,Algebraic=true) 
     catch:
     error "MAPLE NEUMI VYRESIT PRESNOU MINIMALIZACI V %1. KROKU; MINIMALIZOVANA FUNKCE: %2", i, FunkceMinSearch 
    end try;
  else Krok := identify((subs(indets(MinSearch[2], `=`), alpha)),all=false,Rational=true,Algebraic=true);     
 end if;

 for j to P do
  newx0[j]:=x0[j]-Krok*VycisleniDerivace[j];
 end do;

 for j to P do
  x0[j]:=newx0[j];
 end do;

 Iterace[i]:=convert(x0,list);

 for j to P do
  b[j]:=x[j]=x0[j]; 
 end do;

 HodnotaPromennych:=convert(b,list);

 FunkcniHodnota[i]:=eval(f,HodnotaPromennych);

 if VypsatIteracePodrobne
  then print(k=i, (FunkceMinSearch), Krok, -PositionVector(Smer), PositionVector(Iterace[i]), FunkcniHodnota[i]); 
 end if; 

 if VypsatIterace 
  then print(k=i, Krok, -PositionVector(Smer), PositionVector(Iterace[i]), FunkcniHodnota[i]); 
 end if;

 if eval(sum(Derivace[k]*Derivace[k],k=1..P),HodnotaPromennych)=0 
  then print("NALEZLI JSME PRESNE RESENI")
 else end if;

end do;

print(k=i-1,Iterace[i-1],FunkcniHodnota[i-1]);

end proc:

######################
######################

NM := proc (f, P, N, VypsatIterace,VypsatIteracePodrobne)

description "NM=Newtonova metoda, obecna syntaxe NM(f, P, N, VypsatIterace,VypsatIteracePodrobne), f..funkce, P..pocet promennych (ty musi byt zapsany jako x[1], x[2],...), N..pocet iteraci, VypsatIterace=true/false, VypsatIteracePodrobne=true/false";

local x0, i, j, newx0, Promenne, HodnotaPromennych, a, b, c, Iterace, VychoziHodnota, FunkcniHodnota, HessovaMatice, Grd, GradientVycisleni, HessovaMaticeVycisleni, Derivace, Smer,VycisleniDerivace;

for i to P do
 x0[i]:=VychoziBod[i];
end do;

Iterace[0]:=convert(x0,list);

for i to P do
 a[i]:=x[i];
 b[i]:=x[i]=x0[i]; 
end do;

Promenne:=convert(a,list);
HodnotaPromennych:=convert(b,list);

VychoziHodnota:=eval(f,HodnotaPromennych);

for i to P do 
 Derivace[i]:=diff(f,x[i]);
end do; 

Grd:=Gradient(f,Promenne);

HessovaMatice:=Hessian(f,Promenne);

print('__________________________________________');
print("FUNKCE", f); 
print("PRESNE RESENI", minimize(f,location)); 
print("NEWTONOVA METODA");
print('__________________________________________');

print("VYCHOZI BOD", HodnotaPromennych,"S HODNOTOU",simplify(VychoziHodnota));
print('__________________________________________');

print('GRADIENT',PositionVector(convert(Derivace,list))); 
print('HESSOVA_MATICE',HessovaMatice); 
print('__________________________________________');

if VypsatIterace then 
 print("k", "[X_k=X_(k-1)-(INVERZE HESSOVY MATICE V X_(k-1))*(GRADIENT V X_(k-1))]",               "F(X_k)","GRADIENT V X_(k-1)", "INVERZE HESSOVY MATICE V X_(k-1)");
end if;

if VypsatIteracePodrobne then 
 print("k", "[X_k=X_(k-1)-(INVERZE HESSOVY MATICE V X_(k-1))*(GRADIENT V X_(k-1))]", "F(X_k)",    "GRADIENT V X_(k-1)", "HESSOVA MATICE V X_(k-1)", "INVERZE HESSOVY MATICE V X_(k-1)", "SOUCIN INVERZE HESSOVY MATICE A GRADIENTU V X_(k-1)");
end if;

print('__________________________________________');

for i to N while evalf(VectorNorm(eval(Grd,HodnotaPromennych),2))>0 do

 for j to P do
  c[j]:=x0[j];
  VycisleniDerivace[j]:=eval(Derivace[j],HodnotaPromennych);
 end do;

 Iterace:=convert(c,list);
 Smer:=convert(VycisleniDerivace,list);

 HessovaMaticeVycisleni:=eval(HessovaMatice,HodnotaPromennych);
 GradientVycisleni:=eval(Grd,HodnotaPromennych);
 newx0:=identify(VectorAdd(Vector(Iterace), MatrixVectorMultiply(MatrixInverse(HessovaMaticeVycisleni),GradientVycisleni),1,-1),all=false,Rational=true,Algebraic=true); 

 for j to P do
  b[j]:=x[j]=newx0[j]; 
  x0[j]:=newx0[j];
 end do;

 HodnotaPromennych:=convert(b,list);

 if VypsatIteracePodrobne then
  print(k=i, newx0, eval(f,HodnotaPromennych), PositionVector(Smer), HessovaMaticeVycisleni, MatrixInverse(HessovaMaticeVycisleni), MatrixVectorMultiply(MatrixInverse(HessovaMaticeVycisleni),GradientVycisleni));
 end if;

if VypsatIterace then
 print(k=i, newx0, eval(f,HodnotaPromennych), PositionVector(Smer), MatrixInverse(HessovaMaticeVycisleni));
else end if;

od;

if simplify(VectorNorm(eval(Grd,HodnotaPromennych),2))=0 
  then print("NALEZLI JSME PRESNE RESENI")
else end if;

print(k=i-1,newx0,eval(f,HodnotaPromennych))

end proc:

######################
######################

MSG := proc (f, P, N, VypsatIterace,VypsatIteracePodrobne)

description "MSG=Metoda sdruzenych gradientu bez resetu beta, obecna syntaxe MSG(f, P, N, VypsatIterace,VypsatIteracePodrobne), f..funkce, P..pocet promennych (ty musi byt zapsany jako x[1], x[2],...), N..pocet iteraci, VypsatIterace=true/false, VypsatIteracePodrobne=true/false";

local x0, i, j, newx0, Promenne, HodnotaPromennych, a, b, c, d, e, Grd, Derivace, beta, Iterace, VychoziHodnota, FunkcniHodnota, GradientVycisleni, ArgMinSearch, VycisleniDerivace, SlozkySmerovehoVektoru, SlozkyGradientu, Smer, FunkceMinSearch, MinSearch, Krok ;

for i to P do
 x0[i]:=VychoziBod[i];
end do;

Iterace[0]:=convert(x0,list);

for i to P do
 a[i]:=x[i];
 b[i]:=x[i]=x0[i]; 
end do;

Promenne:=convert(a,list);
HodnotaPromennych[0]:=convert(b,list);

VychoziHodnota:=eval(f,HodnotaPromennych[0]);

for i to P do 
 Derivace[i]:=diff(f,x[i]);
end do;

Grd:=Gradient(f,Promenne);
GradientVycisleni[0]:=eval(Grd,HodnotaPromennych[0]);

print('__________________________________________');
print("FUNKCE", f); 
print("PRESNE RESENI", minimize(f,location)); 
print("METODA SDRUZENYCH GRADIENTU");
print('__________________________________________');

print("VYCHOZI BOD", HodnotaPromennych[0] ,"S HODNOTOU", simplify(VychoziHodnota));
print('__________________________________________');

print('GRADIENT',PositionVector(convert(Derivace,list))); 
print('__________________________________________');

if VypsatIterace then 
 print("k", "ALPHA_(k-1)", "[VEKTOR h_(k-1)]", "BETA_(k-1)", "[X_k=X_(k-1)+ALPHA_(k-1)*{h_(k-1)+BETA_(k-2)*h_(k-2)}]", "F(X_k)");
end if;

if VypsatIteracePodrobne then 
 print("k", "GRADIENT V X_(k-1)", "FUNKCE PRO VYPOCET ALPHA", "ALPHA_(k-1)", "[VEKTOR h_(k-1)]", "BETA_(k-1)", "[X_k=X_(k-1)+ALPHA_(k-1)*{h_(k-1)+BETA_(k-2)*h_(k-2)}]", "F(X_k)");
end if;

print('__________________________________________');

beta[-1]:=0;

for i from 1 to N while evalf(VectorNorm(eval(Grd,HodnotaPromennych[i-1]),2))>0 do

 for j to P do
  VycisleniDerivace[i-1][j]:=eval(Derivace[j],HodnotaPromennych[i-1]);
  c[j]:=x[j]=x0[j]+alpha*(-VycisleniDerivace[i-1][j]+beta[i-2]*Smer[i-2][j]);
  SlozkySmerovehoVektoru[i-1][j]:=-VycisleniDerivace[i-1][j]+beta[i-2]*Smer[i-2][j];
 end do;

 ArgMinSearch:=convert(c,list);
 SlozkyGradientu[i-1]:=convert(VycisleniDerivace[i-1],list);
 Smer[i-1]:=convert(SlozkySmerovehoVektoru[i-1],list);

 FunkceMinSearch:=eval(f,ArgMinSearch);

 MinSearch:=minimize(FunkceMinSearch,location):

 if whattype(MinSearch[1])=function 
  then 
    try
    	 MinSearch:=Minimize(FunkceMinSearch):
    	 Krok := identify((subs(indets(MinSearch[2], `=`), alpha)),all=false,Rational=true,Algebraic=true) 
     catch:
     error "MAPLE NEUMI VYRESIT PRESNOU MINIMALIZACI V %1. KROKU; MINIMALIZOVANA FUNKCE: %2", i, FunkceMinSearch 
    end try;
  else Krok := identify((subs(indets(MinSearch[2], `=`), alpha)),all=false,Rational=true,Algebraic=true);     
 end if; 

 for j to P do
  newx0[j]:=x0[j]+Krok*(-VycisleniDerivace[i-1][j]+beta[i-2]*Smer[i-2][j]);
 end do;

 for j to P do
  d[j]:=x[j]=x0[j];
  e[j]:=x[j]=newx0[j];  
 end do;
 
 HodnotaPromennych[i]:=convert(e,list);
 GradientVycisleni[i]:=eval(Grd,HodnotaPromennych[i]);

 beta[i-1]:=identify((VectorNorm(GradientVycisleni[i],2)/VectorNorm(GradientVycisleni[i-1],2))^2,all=false,Rational=true,Algebraic=true);

 for j to P do
  x0[j]:=newx0[j];
 end do;

 Iterace[i]:=convert(x0,list);
 FunkcniHodnota[i]:=eval(f,HodnotaPromennych[i]);

 if VypsatIteracePodrobne then
  print(k=i, PositionVector(SlozkyGradientu[i-1]), simplify(FunkceMinSearch), Krok,    PositionVector(Smer[i-1]), beta[i-1], PositionVector(Iterace[i]), FunkcniHodnota[i]);
 end if;

 if VypsatIterace then
  print(k=i, Krok, PositionVector(Smer[i-1]), beta[i-1], PositionVector(Iterace[i]), FunkcniHodnota[i]);
 end if;

od;

if VectorNorm(eval(Grd,HodnotaPromennych[i-1]),2)=0 
 then print("NALEZLI JSME PRESNE RESENI")
end if;

print(k=i-1,Iterace[i-1],eval(f,HodnotaPromennych[i-1]))

end proc:

######################
######################

MSGreset := proc (f, P, N, VypsatIterace,VypsatIteracePodrobne)

description "MSGreset=Metoda sdruzenych gradientu s resetem beta, obecna syntaxe MSG(f, P, N, VypsatIterace,VypsatIteracePodrobne), f..funkce, P..pocet promennych (ty musi byt zapsany jako x[1], x[2],...), N..pocet iteraci, VypsatIterace=true/false, VypsatIteracePodrobne=true/false";

local x0, i, j, newx0, Promenne, HodnotaPromennych, a, b, c, d, e, Grd, Derivace, beta, Iterace, VychoziHodnota, FunkcniHodnota, GradientVycisleni, ArgMinSearch, VycisleniDerivace, SlozkySmerovehoVektoru, SlozkyGradientu, Smer, FunkceMinSearch, MinSearch, Krok ;

for i to P do
 x0[i]:=VychoziBod[i];
end do;

Iterace[0]:=convert(x0,list);

for i to P do
 a[i]:=x[i];
 b[i]:=x[i]=x0[i]; 
end do;

Promenne:=convert(a,list);
HodnotaPromennych[0]:=convert(b,list);

VychoziHodnota:=eval(f,HodnotaPromennych[0]);

for i to P do 
 Derivace[i]:=diff(f,x[i]);
end do;

Grd:=Gradient(f,Promenne);
GradientVycisleni[0]:=eval(Grd,HodnotaPromennych[0]);

print('__________________________________________');
print("FUNKCE", f); 
print("PRESNE RESENI", minimize(f,location)); 
print("METODA SDRUZENYCH GRADIENTU (s resetovanim beta)");
print('__________________________________________');

print("VYCHOZI BOD", HodnotaPromennych[0] ,"S HODNOTOU", simplify(VychoziHodnota));
print('__________________________________________');

print('GRADIENT',PositionVector(convert(Derivace,list))); 
print('__________________________________________');

if VypsatIterace then 
 print("k", "ALPHA_(k-1)", "[VEKTOR h_(k-1)]", "BETA_(k-1)", "[X_k=X_(k-1)+ALPHA_(k-1)*{h_(k-1)+BETA_(k-2)*h_(k-2)}]", "F(X_k)");
end if;

if VypsatIteracePodrobne then 
 print("k", "GRADIENT V X_(k-1)", "FUNKCE PRO VYPOCET ALPHA", "ALPHA_(k-1)", "[VEKTOR h_(k-1)]", "BETA_(k-1)", "[X_k=X_(k-1)+ALPHA_(k-1)*{h_(k-1)+BETA_(k-2)*h_(k-2)}]", "F(X_k)");
end if;

print('__________________________________________');

beta[-1]:=0;

for i from 1 to N while evalf(VectorNorm(eval(Grd,HodnotaPromennych[i-1]),2))>0 do

 for j to P do
  VycisleniDerivace[i-1][j]:=eval(Derivace[j],HodnotaPromennych[i-1]);
  c[j]:=x[j]=x0[j]+alpha*(-VycisleniDerivace[i-1][j]+beta[i-2]*Smer[i-2][j]);
  SlozkySmerovehoVektoru[i-1][j]:=-VycisleniDerivace[i-1][j]+beta[i-2]*Smer[i-2][j];
 end do;

 ArgMinSearch:=convert(c,list);
 SlozkyGradientu[i-1]:=convert(VycisleniDerivace[i-1],list);
 Smer[i-1]:=convert(SlozkySmerovehoVektoru[i-1],list);

 FunkceMinSearch:=eval(f,ArgMinSearch);

 MinSearch:=minimize(FunkceMinSearch,location):

 if whattype(MinSearch[1])=function 
  then 
    try
    	 MinSearch:=Minimize(FunkceMinSearch):
    	 Krok := identify((subs(indets(MinSearch[2], `=`), alpha)),all=false,Rational=true,Algebraic=true) 
     catch:
     error "MAPLE NEUMI VYRESIT PRESNOU MINIMALIZACI V %1. KROKU; MINIMALIZOVANA FUNKCE: %2", i, FunkceMinSearch 
    end try;
  else Krok := identify((subs(indets(MinSearch[2], `=`), alpha)),all=false,Rational=true,Algebraic=true);     
 end if;
 for j to P do
  newx0[j]:=x0[j]+Krok*(-VycisleniDerivace[i-1][j]+beta[i-2]*Smer[i-2][j]);
 end do;

 for j to P do
  d[j]:=x[j]=x0[j];
  e[j]:=x[j]=newx0[j];  
 end do;
 
 HodnotaPromennych[i]:=convert(e,list);
 GradientVycisleni[i]:=eval(Grd,HodnotaPromennych[i]);

 if modp(i,P)=0 then 
  beta[i-1]:=0
 else    
  beta[i-1]:=identify((VectorNorm(GradientVycisleni[i],2)/VectorNorm(GradientVycisleni[i-1],2))^2,all=false,Rational=true,Algebraic=true);
 end if;

 for j to P do
  x0[j]:=newx0[j];
 end do;

 Iterace[i]:=convert(x0,list);
 FunkcniHodnota[i]:=eval(f,HodnotaPromennych[i]);

 if VypsatIteracePodrobne then
  print(k=i, PositionVector(SlozkyGradientu[i-1]), simplify(FunkceMinSearch), Krok,    PositionVector(Smer[i-1]), beta[i-1], PositionVector(Iterace[i]), FunkcniHodnota[i]);
 end if;

 if VypsatIterace then
  print(k=i, Krok, PositionVector(Smer[i-1]), beta[i-1], PositionVector(Iterace[i]), FunkcniHodnota[i]);
 end if;

od;

if VectorNorm(eval(Grd,HodnotaPromennych[i-1]),2)=0 
 then print("NALEZLI JSME PRESNE RESENI")
end if;

print(k=i-1,Iterace[i-1],eval(f,HodnotaPromennych[i-1]))

end proc:

end module;# 
;
